## hatch_ci.tools

src/hatch_ci/tools.py
    1
    2
    3
    4
    5
    6
    7
    8
    9
    10
    11
    12
    13
    14
    15
    16
    17
    18
    19
    20
    21
    22
    23
    24
    25
    26
    27
    28
    29
    30
    31
    32
    33
    34
    35
    36
    37
    38
    39
    40
    41
    42
    43
    44
    45
    46
    47
    48
    49
    50
    51
    52
    53
    54
    55
    56
    57
    58
    59
    60
    61
    62
    63
    64
    65
    66
    67
    68
    69
    70
    71
    72
    73
    74
    75
    76
    77
    78
    79
    80
    81
    82
    83
    84
    85
    86
    87
    88
    89
    90
    91
    92
    93
    94
    95
    96
    97
    98
    99
    100
    101
    102
    103
    104
    105
    106
    107
    108
    109
    110
    111
    112
    113
    114
    115
    116
    117
    118
    119
    120
    121
    122
    123
    124
    125
    126
    127
    128
    129
    130
    131
    132
    133
    134
    135
    136
    137
    138
    139
    140
    141
    142
    143
    144
    145
    146
    147
    148
    149
    150
    151
    152
    153
    154
    155
    156
    157
    158
    159
    160
    161
    162
    163
    164
    165
    166
    167
    168
    169
    170
    171
    172
    173
    174
    175
    176
    177
    178
    179
    180
    181
    182
    183
    184
    185
    186
    187
    188
    189
    190
    191
    192
    193
    194
    195
    196
    197
    198
    199
    200
    201
    202
    203
    204
    205
    206
    207
    208
    209
    210
    211
    212
    213
    214
    215
    216
    217
    218
    219
    220
    221
    222
    223
    224
    225
    226
    227
    228
    229
    230
    231
    232
    233
    234
    235
    236
    237
    238
    239
    240
    241
    242
    243
    244
    245
    246
    247
    248
    249
    250
    251
    252
    253
    254
    255
    256
    257
    258
    259
    260
    261
    262
    263
    264
    265
    266
    267
    268
    269
    270
    271
    272
    273
    274
    275
    276
    277
    278
    279
    280
    281
    from __future__ import annotations

    import ast
    import json
    import re
    from pathlib import Path
    from typing import Any

    from . import scm

    class ToolsError(Exception):
        pass

    class ValidationError(ToolsError):
        pass

    class InvalidVersionError(ToolsError):
        pass

    class MissingVariableError(ToolsError):
        pass

    class AbortExecutionError(Exception):
        @staticmethod
        def _strip(txt):
            txt = txt or ""
            txt = txt[1:] if txt.startswith("\n") else txt
            txt = indent(txt, pre="")
            return txt[:-1] if txt.endswith("\n") else txt

        def __init__(
            self, message: str, explain: str | None = None, hint: str | None = None
        ):
            self.message = message.strip()
            self._explain = explain
            self._hint = hint

        @property
        def explain(self):
            return self._strip(self._explain)

        @property
        def hint(self):
            return self._strip(self._hint)

        def __str__(self):
            result = [self.message]
            if self.explain:
                result.append(indent("\n" + self.explain, pre=" " * 2)[2:])
            if self.hint:
                result.extend(["\nhint:", indent("\n" + self.hint, pre=" " * 2)[2:]])
            return "".join(result)

    def urmtree(path: Path):
        "universal (win|*nix) rmtree"
        from os import name
        from shutil import rmtree
        from stat import S_IWUSR

        if name == "nt":
            for p in path.rglob("*"):
                p.chmod(S_IWUSR)
        rmtree(path, ignore_errors=True)
        if path.exists():
            raise RuntimeError(f"cannot remove {path=}")

    def indent(txt: str, pre: str = " " * 2) -> str:
        "simple text indentation"

        from textwrap import dedent

        txt = dedent(txt)
        if txt.endswith("\n"):
            last_eol = "\n"
            txt = txt[:-1]
        else:
            last_eol = ""

        result = pre + txt.replace("\n", "\n" + pre) + last_eol
        return result if result.strip() else result.strip()

    def list_of_paths(paths: str | Path | list[str | Path]) -> list[Path]:
        return [Path(s) for s in ([paths] if isinstance(paths, (str, Path)) else paths)]

    def get_module_var(
        path: Path | str, var: str = "__version__", abort=True
    ) -> str | None:
        """extract from a python module in path the module level <var> variable

        Args:
            path (str,Path): python module file to parse using ast (no code-execution)
            var (str): module level variable name to extract
            abort (bool): raise MissingVariable if var is not present

        Returns:
            None or str: the variable value if found or None

        Raises:
            MissingVariable: if the var is not found and abort is True

        Notes:
            this uses ast to parse path, so it doesn't load the module
        """

        class V(ast.NodeVisitor):
            def __init__(self, keys):
                self.keys = keys
                self.result = {}

            def visit_Module(self, node):  # noqa: N802
                # we extract the module level variables
                for subnode in ast.iter_child_nodes(node):
                    if not isinstance(subnode, ast.Assign):
                        continue
                    for target in subnode.targets:
                        if target.id not in self.keys:
                            continue
                        if not isinstance(subnode.value, (ast.Num, ast.Str, ast.Constant)):
                            raise ValidationError(
                                f"cannot extract non Constant variable "
                                f"{target.id} ({type(subnode.value)})"
                            )
                        if isinstance(subnode.value, ast.Str):
                            value = subnode.value.s
                        elif isinstance(subnode.value, ast.Num):
                            value = subnode.value.n
                        else:
                            value = subnode.value.value
                        self.result[target.id] = value
                return self.generic_visit(node)

        v = V({var})
        path = Path(path)
        if path.exists():
            tree = ast.parse(Path(path).read_text())
            v.visit(tree)
        if var not in v.result and abort:
            raise MissingVariableError(f"cannot find {var} in {path}", path, var)
        return v.result.get(var, None)

    def set_module_var(
        path: str | Path, var: str, value: Any, create: bool = True
    ) -> tuple[Any, str]:
        """replace var in path with value

        Args:
            path (str,Path): python module file to parse
            var (str): module level variable name to extract
            value (None or Any): if not None replace var in initfile
            create (bool): create path if not present

        Returns:
            (str, str) the (<previous-var-value|None>, <the new text>)
        """
        # module level var
        expr = re.compile(f"^{var}\\s*=\\s*['\\\"](?P<value>[^\\\"']*)['\\\"]")
        fixed = None
        lines = []

        src = Path(path)
        if not src.exists() and create:
            src.parent.mkdir(parents=True, exist_ok=True)
            src.touch()

        input_lines = src.read_text().split("\n")
        for line in reversed(input_lines):
            if fixed:
                lines.append(line)
                continue
            match = expr.search(line)
            if match:
                fixed = match.group("value")
                if value is not None:
                    x, y = match.span(1)
                    line = line[:x] + value + line[y:]
            lines.append(line)
        txt = "\n".join(reversed(lines))
        if not fixed and create:
            if txt and txt[-1] != "\n":
                txt += "\n"
            txt += f'{var} = "{value}"'

        with Path(path).open("w") as fp:
            fp.write(txt)
        return fixed, txt

    def bump_version(version: str, mode: str) -> str:
        """given a version str will bump it according to mode

        Arguments:
            version: text in the N.M.O form
            mode: major, minor or micro

        Returns:
            increased text

        >>> bump_version("1.0.3", "micro")
        "1.0.4"
        >>> bump_version("1.0.3", "minor")
        "1.1.0"
        """
        newver = [int(n) for n in version.split(".")]
        if mode == "major":
            newver[-3] += 1
            newver[-2] = 0
            newver[-1] = 0
        elif mode == "minor":
            newver[-2] += 1
            newver[-1] = 0
        elif mode == "micro":
            newver[-1] += 1
        return ".".join(str(v) for v in newver)

    def update_version(
        initfile: str | Path, github_dump: str | None = None, abort: bool = True
    ) -> str | None:
        """extracts version information from github_dump and updates initfile in-place

        Args:
            initfile (str, Path): path to the __init__.py file with a __version__ variable
            github_dump (str): the os.getenv("GITHUB_DUMP") value

        Returns:
            str: the new version for the package
        """

        path = Path(initfile)
        repo = scm.lookup(path)

        if not (repo or github_dump):
            if abort:
                raise scm.InvalidGitRepoError(f"cannot find a valid git repo for {path}")
            return get_module_var(path, "__version__")

        if not github_dump and repo:
            gdata = {
                "ref": repo.head.name,
                "sha": repo.head.target.hex[:7],
                "run_number": 0,
            }
            dirty = bool(repo.status())
        else:
            gdata = json.loads(github_dump) if isinstance(github_dump, str) else github_dump
            dirty = False

        version = current = get_module_var(path, "__version__")

        expr = re.compile(r"/(?P<what>beta|release)/(?P<version>\d+([.]\d+)*)$")
        expr1 = re.compile(r"(?P<version>\d+([.]\d+)*)(?P<num>b\d+)?$")

        if match := expr.search(gdata["ref"]):
            # setuptools double calls the update_version,
            # this fixes the issue
            match1 = expr1.search(current or "")
            if not match1:
                raise InvalidVersionError(f"cannot parse current version '{current}'")
            if match1.group("version") != match.group("version"):
                raise InvalidVersionError(
                    f"building package for {current} from '{gdata['ref']}' "
                    f"branch ({match.groupdict()} mismatch {match1.groupdict()})"
                )
            if match.group("what") == "beta":
                version = f"{match1.group('version')}b{gdata['run_number']}"

        short = gdata["sha"] + ("*" if dirty else "")

        set_module_var(path, "__version__", version)
        set_module_var(path, "__hash__", short)
        return version